#import <Cocoa/Cocoa.h>

#define XeeNoMode 0
#define XeeMoveMode 1
#define XeeCopyMode 2

@class XeeImage,XeeImageSource,XeeView,XeeDisplayWindow,XeeFullScreenWindow;
@class XeeCollisionPanel,XeeRenamePanel,XeeStatusBar,XeeDestinationView;



@interface XeeController:NSObject
{
	XeeImageSource *source;
	XeeImage *currimage;

	float zoom;
	int window_focus_x,window_focus_y;
	BOOL blocked;

	float scrollpos;
	int drawer_mode;

	NSDictionary *toolbaritems;
	NSArray *toolbaridentifiers;

	NSUndoManager *undo;

	XeeFullScreenWindow *fullscreenwindow;
	XeeView *fullscreenview;

	NSTimer *slideshowtimer;
	int slideshowcount;

	CGImageRef copiedcgimage;

	BOOL awake;

	IBOutlet NSWindow *window;
    IBOutlet XeeView *imageview;
    IBOutlet XeeStatusBar *statusbar;
    IBOutlet NSDrawer *drawer;
	IBOutlet NSSegmentedControl *drawerseg;
	IBOutlet XeeDestinationView *destinationtable;
	IBOutlet NSButton *closebutton;
	IBOutlet XeeCollisionPanel *collisionpanel;
	IBOutlet XeeRenamePanel *renamepanel;
	IBOutlet NSTextField *delayfield;
	IBOutlet NSPanel *delaypanel;
}

-(id)init;
-(void)dealloc;
-(void)awakeFromNib;

-(void)windowWillClose:(NSNotification *)notification;
-(void)windowDidBecomeMain:(NSNotification *)notification;
-(void)windowDidResignMain:(NSNotification *)notification;
-(void)windowDidMove:(NSNotification *)notification;
-(void)windowDidResize:(NSNotification *)notification;
-(void)windowWillMiniaturize:(NSNotification *)notification;
-(void)windowDidMiniaturize:(NSNotification *)notification;
-(NSUndoManager *)windowWillReturnUndoManager:(NSNotification *)notification;
-(void)setStatusBarHiddenNotification:(NSNotification *)notification;

-(void)scrollWheel:(NSEvent *)event;

-(void)xeeImageSource:(XeeImageSource *)msgsource imageDidChange:(XeeImage *)image;
-(void)xeeImageSource:(XeeImageSource *)source imageListDidChange:(int)num;
-(void)xeeView:(XeeView *)view imageDidChange:(XeeImage *)image;
-(void)xeeView:(XeeView *)view imageSizeDidChange:(XeeImage *)image;
-(void)xeeView:(XeeView *)view imagePropertiesDidChange:(XeeImage *)image;

-(NSWindow *)window;
-(XeeFullScreenWindow *)fullScreenWindow;
-(XeeImageSource *)imageSource;
-(XeeImage *)image;
-(NSDrawer *)drawer;
-(NSString *)currentFilename;
-(BOOL)isFullscreen;

-(void)setImageSource:(XeeImageSource *)newsource;
-(void)setImage:(XeeImage *)image;
-(void)setZoom:(float)newzoom;
-(void)setFrame:(int)frame;

-(void)updateWindowPosition;
-(void)setImageSize:(NSSize)size;
-(void)setImageSize:(NSSize)size resetFocus:(BOOL)reset;
-(void)setStandardImageSize;
-(void)setResizeBlock:(BOOL)block;
-(void)setResizeBlockFromSender:(id)sender;
-(BOOL)isResizeBlocked;
-(NSSize)maxViewSize;
-(NSSize)minViewSize;
-(NSRect)availableScreenSpace;

-(void)errorMessage:(NSString *)title text:(NSString *)text;

-(NSToolbarItem *)toolbar:(NSToolbar *)toolbar itemForItemIdentifier:(NSString *)identifier willBeInsertedIntoToolbar:(BOOL)flag;
-(NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar *)toolbar;
-(void)setupToolbarItems;
-(NSArray *)makeToolbarItems;

-(BOOL)validateMenuItem:(id <NSMenuItem>)item;
-(BOOL)validateAction:(SEL)action;

-(void)updateStatusBar;
-(void)setStatusBarHidden:(BOOL)hidden;
-(BOOL)isStatusBarHidden;
-(IBAction)toggleStatusBar:(id)sender;

-(void)setDrawerEnableState;

-(IBAction)fullScreen:(id)sender;

-(IBAction)closeWindowOrDrawer:(id)sender;


+(void)initialize;
+(NSArray *)controllers;

@end




@interface XeeFullScreenWindow:NSWindow
{
}

-(BOOL)canBecomeKeyWindow;

@end
