#import "XeeControllerImageActions.h"
#import "XeeSavePanel.h"
#import "XeeImageSaver.h"



float XeeZoomLevels[]={0.03125,0.044,0.0625,0.09,0.125,0.18,0.25,0.35,0.5,0.70,1,1.5,2,3,4,6,8,11,16,23,32};
int XeeNumberOfZoomLevels=21;



@implementation XeeController (ImageActions)

-(IBAction)copy:(id)sender
{
	if(currimage)
	{
		CGImageRef cgimage=[currimage createCGImage];

		if(cgimage)
		{
			[[NSPasteboard generalPasteboard] declareTypes:[NSArray arrayWithObjects:NSTIFFPboardType,NSPICTPboardType,nil] owner:self];

			copiedcgimage=cgimage;
			[self retain];
		}
		else NSBeep();
	}
	else NSBeep();
}

-(void)pasteboard:(NSPasteboard *)pboard provideDataForType:(NSString *)type
{
	if(!copiedcgimage) return;

	CFStringRef uttype;
	if([type isEqual:NSTIFFPboardType]) uttype=kUTTypeTIFF;
	else uttype=kUTTypePICT;

	NSMutableData *data=[NSMutableData data];

	CGImageDestinationRef dest=CGImageDestinationCreateWithData((CFMutableDataRef)data,uttype,1,NULL);
	if(!dest) { NSBeep(); return; }

	CGImageDestinationAddImage(dest,copiedcgimage,NULL);
	if(!CGImageDestinationFinalize(dest)) NSBeep();
	CFRelease(dest);

	[pboard setData:data forType:type];
}

-(void)pasteboardChangedOwner:(NSPasteboard *)pboard
{
	CGImageRelease(copiedcgimage);
	copiedcgimage=NULL;

	[self release];
}



-(IBAction)saveImage:(id)sender
{
	if(currimage) [XeeSavePanel runSavePanelForImage:currimage window:fullscreenwindow?nil:window];
}



-(IBAction)frameSkipNext:(id)sender
{
	[self setResizeBlockFromSender:sender];
	if(currimage)
	{
		int frame=[currimage frame];
		int frames=[currimage frames];
		[self setFrame:(frame+1)%frames];
	}
	[self setResizeBlock:NO];
}

-(IBAction)frameSkipPrev:(id)sender
{
	[self setResizeBlockFromSender:sender];
	if(currimage)
	{
		int frame=[currimage frame];
		int frames=[currimage frames];
		[self setFrame:(frame+frames-1)%frames];
	}
	[self setResizeBlock:NO];
}

-(IBAction)toggleAnimation:(id)sender
{
	if(!currimage||![currimage animated]) return;

	[currimage setAnimating:![currimage animating]];
}



-(IBAction)zoomIn:(id)sender
{
	[self setResizeBlockFromSender:sender];

	int i;
	for(i=0;i<XeeNumberOfZoomLevels-1;i++) if(XeeZoomLevels[i]>zoom) break;

	[self setZoom:XeeZoomLevels[i]];

	[self setResizeBlock:NO];
}

-(IBAction)zoomOut:(id)sender
{
	[self setResizeBlockFromSender:sender];

	int i;
	for(i=XeeNumberOfZoomLevels-1;i>0;i--) if(XeeZoomLevels[i]<zoom) break;

	[self setZoom:XeeZoomLevels[i]];

	[self setResizeBlock:NO];
}

-(IBAction)zoomActual:(id)sender
{
	[self setResizeBlockFromSender:sender];
	[self setZoom:1];
	[self setResizeBlock:NO];
}

-(IBAction)zoomFit:(id)sender
{
	[self setResizeBlockFromSender:sender];

	NSSize maxsize=[self maxViewSize];

	float horiz_zoom=maxsize.width/(float)[currimage width];
	float vert_zoom=maxsize.height/(float)[currimage height];

	[self setZoom:horiz_zoom<vert_zoom?horiz_zoom:vert_zoom];

	[self setResizeBlock:NO];
}

-(IBAction)setAutoZoom:(id)sender
{
	[self setResizeBlockFromSender:sender];
	[self setStandardImageSize];
	[self setResizeBlock:NO];
}

-(IBAction)setAntialiasing:(id)sender
{
	[[NSUserDefaults standardUserDefaults] setInteger:[sender tag] forKey:@"antialiasQuality"];

	NSMenu *menu=[sender menu];
	int num=[menu numberOfItems];
	for(int i=0;i<num;i++)
	{
		NSMenuItem *item=[menu itemAtIndex:i];
		[item setState:item==sender?NSOnState:NSOffState];
	}

	[imageview invalidate];
}



-(void)setOrientation:(int)orientation
{
	[[undo prepareWithInvocationTarget:self] setOrientation:[currimage orientation]];
	[currimage setOrientation:orientation];
}

-(IBAction)rotateCW:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setOrientation:XeeCombineTransformations([currimage orientation],XeeRotateCWTransformation)];
	[undo setActionName:NSLocalizedString(@"Rotate Clockwise",@"Name of rotate clockwise action in undo list")];
	[self setResizeBlock:NO];
}

-(IBAction)rotateCCW:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setOrientation:XeeCombineTransformations([currimage orientation],XeeRotateCCWTransformation)];
	[undo setActionName:NSLocalizedString(@"Rotate Counter-clockwise",@"Name of rotate counter-clockwise action in undo list")];
	[self setResizeBlock:NO];
}

-(IBAction)rotate180:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setOrientation:XeeCombineTransformations([currimage orientation],XeeRotate180Transformation)];
	[undo setActionName:NSLocalizedString(@"Rotate 180º",@"Name of rotate 180 action in undo list")];
	[self setResizeBlock:NO];
}

-(IBAction)autoRotate:(id)sender
{
	if(!currimage) return;
	XeeTransformation orientation=[currimage correctOrientation];
	if(orientation==XeeUnknownTransformation) return;

	[self setResizeBlockFromSender:sender];
	[self setOrientation:orientation];
	[undo setActionName:NSLocalizedString(@"Automatic Orientation",@"Name of automatic orientation action in undo list")];
	[self setResizeBlock:NO];
}

-(IBAction)rotateActual:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setOrientation:XeeNoTransformation];
	[undo setActionName:NSLocalizedString(@"Actual Orientation",@"Name of actual orientation action in undo list")];
	[self setResizeBlock:NO];
}

-(IBAction)mirrorHorizontal:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setOrientation:XeeCombineTransformations([currimage orientation],XeeMirrorHorizontalTransformation)];
	[undo setActionName:NSLocalizedString(@"Mirror Horizontal",@"Name of mirror horizontal action in undo list")];
	[self setResizeBlock:NO];
}

-(IBAction)mirrorVertical:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setOrientation:XeeCombineTransformations([currimage orientation],XeeMirrorVerticalTransformation)];
	[undo setActionName:NSLocalizedString(@"Mirror Vertical",@"Name of mirror vertical action in undo list")];
	[self setResizeBlock:NO];
}



-(void)setCroppingRect:(NSRect)rect
{
	[[undo prepareWithInvocationTarget:self] setCroppingRect:[currimage croppingRect]];
	[currimage setCroppingRect:rect];
}

-(IBAction)crop:(id)sender
{
	if(!currimage) return;
	[self setResizeBlockFromSender:sender];
	[self setCroppingRect:NSMakeRect([currimage width]/4,[currimage height]/4,[currimage width]/2,[currimage height]/2)];
	[undo setActionName:NSLocalizedString(@"Crop",@"Name of crop action in undo list")];
	[self setResizeBlock:NO];
}


@end
